from __future__ import annotations

# fmt: off
ORDERED_DISTRIBUTIONS: tuple[str, ...] = (
    '3.7',
    '3.8',
    '3.9',
    '3.10',
    '3.11',
    '3.12',
    '3.13',
    '3.14',
    'pypy2.7',
    'pypy3.9',
    'pypy3.10',
    'pypy3.11',
)
DISTRIBUTIONS: dict[str, dict[tuple[str, ...], str]] = {
    '3.14': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'aarch64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'musl', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'armv7', 'gnueabi', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-armv7-unknown-linux-gnueabi-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabi', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-armv7-unknown-linux-gnueabi-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'armv7', 'gnueabihf', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-armv7-unknown-linux-gnueabihf-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabihf', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-armv7-unknown-linux-gnueabihf-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'ppc64le', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-ppc64le-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'ppc64le', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-ppc64le-unknown-linux-gnu-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'riscv64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-riscv64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'riscv64', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-riscv64-unknown-linux-gnu-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 's390x', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-s390x-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 's390x', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-s390x-unknown-linux-gnu-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v2-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v2', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v2-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v3-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v3', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v3-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v4-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v4', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v4-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v2-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v2', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v2-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v3-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v3', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v3-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v4-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v4', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64_v4-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('windows', 'aarch64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'aarch64', 'msvc', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-i686-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-aarch64-apple-darwin-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.14.0%2B20251014-x86_64-apple-darwin-freethreaded%2Bpgo%2Blto-full.tar.zst',
    },
    '3.13': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'aarch64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'musl', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'armv7', 'gnueabi', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-armv7-unknown-linux-gnueabi-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabi', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-armv7-unknown-linux-gnueabi-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'armv7', 'gnueabihf', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-armv7-unknown-linux-gnueabihf-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabihf', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-armv7-unknown-linux-gnueabihf-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'ppc64le', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-ppc64le-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'ppc64le', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-ppc64le-unknown-linux-gnu-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'riscv64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-riscv64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'riscv64', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-riscv64-unknown-linux-gnu-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 's390x', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-s390x-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 's390x', 'gnu', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-s390x-unknown-linux-gnu-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v2-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v2', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v2-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v3-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v3', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v3-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'gnu', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v4-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v4', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v4-unknown-linux-gnu-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v2-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v2', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v2-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v3-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v3', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v3-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('linux', 'x86_64', 'musl', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v4-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v4', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64_v4-unknown-linux-musl-freethreaded%2Bnoopt-full.tar.zst',
        ('windows', 'aarch64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'aarch64', 'msvc', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-i686-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-aarch64-apple-darwin-freethreaded%2Bpgo%2Blto-full.tar.zst',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', 'freethreaded'):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.13.9%2B20251014-x86_64-apple-darwin-freethreaded%2Bpgo%2Blto-full.tar.zst',
    },
    '3.12': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-aarch64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabi', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-armv7-unknown-linux-gnueabi-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabihf', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-armv7-unknown-linux-gnueabihf-install_only_stripped.tar.gz',
        ('linux', 'ppc64le', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-ppc64le-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'riscv64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-riscv64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 's390x', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-s390x-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64_v2-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64_v3-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64_v4-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64_v2-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64_v3-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64_v4-unknown-linux-musl-install_only_stripped.tar.gz',
        ('windows', 'aarch64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-aarch64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.12.12%2B20251014-x86_64-apple-darwin-install_only_stripped.tar.gz',
    },
    '3.11': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-aarch64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabi', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-armv7-unknown-linux-gnueabi-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabihf', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-armv7-unknown-linux-gnueabihf-install_only_stripped.tar.gz',
        ('linux', 'ppc64le', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-ppc64le-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'riscv64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-riscv64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 's390x', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-s390x-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64_v2-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64_v3-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64_v4-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64_v2-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64_v3-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64_v4-unknown-linux-musl-install_only_stripped.tar.gz',
        ('windows', 'aarch64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-aarch64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.11.14%2B20251014-x86_64-apple-darwin-install_only_stripped.tar.gz',
        ('linux', 'i686', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20230826/cpython-3.11.5%2B20230826-i686-unknown-linux-gnu-install_only.tar.gz',
    },
    '3.10': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-aarch64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabi', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-armv7-unknown-linux-gnueabi-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabihf', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-armv7-unknown-linux-gnueabihf-install_only_stripped.tar.gz',
        ('linux', 'ppc64le', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-ppc64le-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'riscv64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-riscv64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 's390x', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-s390x-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64_v2-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64_v3-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64_v4-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64_v2-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64_v3-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64_v4-unknown-linux-musl-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.10.19%2B20251014-x86_64-apple-darwin-install_only_stripped.tar.gz',
        ('linux', 'i686', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20230826/cpython-3.10.13%2B20230826-i686-unknown-linux-gnu-install_only.tar.gz',
    },
    '3.9': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'aarch64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-aarch64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabi', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-armv7-unknown-linux-gnueabi-install_only_stripped.tar.gz',
        ('linux', 'armv7', 'gnueabihf', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-armv7-unknown-linux-gnueabihf-install_only_stripped.tar.gz',
        ('linux', 'ppc64le', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-ppc64le-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'riscv64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-riscv64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 's390x', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-s390x-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64_v2-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64_v3-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64_v4-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v2', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64_v2-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v3', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64_v3-unknown-linux-musl-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v4', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64_v4-unknown-linux-musl-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20251014/cpython-3.9.24%2B20251014-x86_64-apple-darwin-install_only_stripped.tar.gz',
        ('linux', 'i686', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20230826/cpython-3.9.18%2B20230826-i686-unknown-linux-gnu-install_only.tar.gz',
    },
    '3.8': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-aarch64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'gnu', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-x86_64-unknown-linux-gnu-install_only_stripped.tar.gz',
        ('linux', 'x86_64', 'musl', 'v1', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-x86_64-unknown-linux-musl-install_only_stripped.tar.gz',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-i686-pc-windows-msvc-install_only_stripped.tar.gz',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-x86_64-pc-windows-msvc-install_only_stripped.tar.gz',
        ('macos', 'arm64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-aarch64-apple-darwin-install_only_stripped.tar.gz',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20241002/cpython-3.8.20%2B20241002-x86_64-apple-darwin-install_only_stripped.tar.gz',
        ('linux', 'i686', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20230826/cpython-3.8.17%2B20230826-i686-unknown-linux-gnu-install_only.tar.gz',
    },
    '3.7': {
        ('linux', 'x86_64', 'gnu', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20200822/cpython-3.7.9-x86_64-unknown-linux-gnu-pgo-20200823T0036.tar.zst',
        ('linux', 'x86_64', 'musl', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20200822/cpython-3.7.9-x86_64-unknown-linux-musl-noopt-20200823T0036.tar.zst',
        ('windows', 'i386', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20200822/cpython-3.7.9-i686-pc-windows-msvc-shared-pgo-20200823T0159.tar.zst',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20200822/cpython-3.7.9-x86_64-pc-windows-msvc-shared-pgo-20200823T0118.tar.zst',
        ('macos', 'x86_64', '', '', ''):
            'https://github.com/astral-sh/python-build-standalone/releases/download/20200823/cpython-3.7.9-x86_64-apple-darwin-pgo-20200823T2228.tar.zst',
    },
    'pypy3.11': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy3.11-v7.3.20-aarch64.tar.bz2',
        ('linux', 'x86_64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy3.11-v7.3.20-linux64.tar.bz2',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://downloads.python.org/pypy/pypy3.11-v7.3.20-win64.zip',
        ('macos', 'arm64', '', '', ''):
            'https://downloads.python.org/pypy/pypy3.11-v7.3.20-macos_arm64.tar.bz2',
        ('macos', 'x86_64', '', '', ''):
            'https://downloads.python.org/pypy/pypy3.11-v7.3.20-macos_x86_64.tar.bz2',
    },
    'pypy3.10': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy3.10-v7.3.19-aarch64.tar.bz2',
        ('linux', 'x86_64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy3.10-v7.3.19-linux64.tar.bz2',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://downloads.python.org/pypy/pypy3.10-v7.3.19-win64.zip',
        ('macos', 'arm64', '', '', ''):
            'https://downloads.python.org/pypy/pypy3.10-v7.3.19-macos_arm64.tar.bz2',
        ('macos', 'x86_64', '', '', ''):
            'https://downloads.python.org/pypy/pypy3.10-v7.3.19-macos_x86_64.tar.bz2',
    },
    'pypy3.9': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy3.9-v7.3.16-aarch64.tar.bz2',
        ('linux', 'x86_64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy3.9-v7.3.16-linux64.tar.bz2',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://downloads.python.org/pypy/pypy3.9-v7.3.16-win64.zip',
        ('macos', 'arm64', '', '', ''):
            'https://downloads.python.org/pypy/pypy3.9-v7.3.16-macos_arm64.tar.bz2',
        ('macos', 'x86_64', '', '', ''):
            'https://downloads.python.org/pypy/pypy3.9-v7.3.16-macos_x86_64.tar.bz2',
    },
    'pypy2.7': {
        ('linux', 'aarch64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy2.7-v7.3.20-aarch64.tar.bz2',
        ('linux', 'x86_64', 'gnu', '', ''):
            'https://downloads.python.org/pypy/pypy2.7-v7.3.20-linux64.tar.bz2',
        ('windows', 'amd64', 'msvc', '', ''):
            'https://downloads.python.org/pypy/pypy2.7-v7.3.20-win64.zip',
        ('macos', 'arm64', '', '', ''):
            'https://downloads.python.org/pypy/pypy2.7-v7.3.20-macos_arm64.tar.bz2',
        ('macos', 'x86_64', '', '', ''):
            'https://downloads.python.org/pypy/pypy2.7-v7.3.20-macos_x86_64.tar.bz2',
    },
}



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Show the available environments")
@click.argument("envs", required=False, nargs=-1)
@click.option("--ascii", "force_ascii", is_flag=True, help="Whether or not to only use ASCII characters")
@click.option("--json", "as_json", is_flag=True, help="Whether or not to output in JSON format")
@click.option("--internal", "-i", is_flag=True, help="Show internal environments")
@click.option("--hide-titles", is_flag=True, hidden=True)
@click.pass_obj
def show(
    app: Application,
    *,
    envs: tuple[str, ...],
    force_ascii: bool,
    as_json: bool,
    internal: bool,
    hide_titles: bool,
):
    """Show the available environments."""
    app.ensure_environment_plugin_dependencies()

    from hatch.config.constants import AppEnvVars

    if as_json:
        import json

        contextual_config = {}
        for environments in (app.project.config.envs, app.project.config.internal_envs):
            for env_name, config in environments.items():
                environment = app.project.get_environment(env_name)
                new_config = contextual_config[env_name] = dict(config)

                env_vars = dict(environment.env_vars)
                env_vars.pop(AppEnvVars.ENV_ACTIVE)
                if env_vars:
                    new_config["env-vars"] = env_vars

                num_dependencies = len(config.get("dependencies", []))
                dependencies = environment.environment_dependencies[:num_dependencies]
                if dependencies:
                    new_config["dependencies"] = dependencies

                extra_dependencies = environment.environment_dependencies[num_dependencies:]
                if extra_dependencies:
                    new_config["extra-dependencies"] = extra_dependencies

                if environment.pre_install_commands:
                    new_config["pre-install-commands"] = list(
                        environment.resolve_commands(environment.pre_install_commands)
                    )

                if environment.post_install_commands:
                    new_config["post-install-commands"] = list(
                        environment.resolve_commands(environment.post_install_commands)
                    )

                if environment.scripts:
                    new_config["scripts"] = {
                        script: list(environment.resolve_commands([script])) for script in environment.scripts
                    }

        app.display(json.dumps(contextual_config, separators=(",", ":")))
        return

    from hatch.dep.core import Dependency, InvalidDependencyError
    from hatchling.metadata.utils import get_normalized_dependency, normalize_project_name

    if internal:
        target_standalone_envs = app.project.config.internal_envs
        target_matrices = app.project.config.internal_matrices
    else:
        target_standalone_envs = app.project.config.envs
        target_matrices = app.project.config.matrices

    for env_name in envs:
        if env_name not in target_standalone_envs and env_name not in target_matrices:
            app.abort(f"Environment `{env_name}` is not defined by project config")

    env_names = set(envs)

    matrix_columns: dict[str, dict[int, str]] = {
        "Name": {},
        "Type": {},
        "Envs": {},
        "Features": {},
        "Dependencies": {},
        "Environment variables": {},
        "Scripts": {},
        "Description": {},
    }
    matrix_envs = set()
    for i, (matrix_name, matrix_data) in enumerate(target_matrices.items()):
        matrix_envs.update(matrix_data["envs"])

        if env_names and matrix_name not in env_names:
            continue

        config = matrix_data["config"]
        matrix_columns["Name"][i] = matrix_name
        matrix_columns["Type"][i] = config["type"]
        matrix_columns["Envs"][i] = "\n".join(matrix_data["envs"])

        if config.get("features"):
            if app.project.metadata.hatch.metadata.allow_ambiguous_features:
                matrix_columns["Features"][i] = "\n".join(sorted(set(config["features"])))
            else:
                matrix_columns["Features"][i] = "\n".join(
                    sorted({normalize_project_name(f) for f in config["features"]})
                )

        dependencies = []
        if config.get("dependencies"):
            dependencies.extend(config["dependencies"])
        if config.get("extra-dependencies"):
            dependencies.extend(config["extra-dependencies"])
        if dependencies:
            normalized_dependencies = set()
            for dependency in dependencies:
                try:
                    req = Dependency(dependency)
                except InvalidDependencyError:
                    normalized_dependencies.add(dependency)
                else:
                    normalized_dependencies.add(get_normalized_dependency(req))

            matrix_columns["Dependencies"][i] = "\n".join(sorted(normalized_dependencies))

        if config.get("env-vars"):
            matrix_columns["Environment variables"][i] = "\n".join(
                "=".join(item) for item in sorted(config["env-vars"].items())
            )

        if config.get("scripts"):
            matrix_columns["Scripts"][i] = "\n".join(
                sorted(script for script in config["scripts"] if app.verbose or not script.startswith("_"))
            )

        if config.get("description"):
            matrix_columns["Description"][i] = config["description"].strip()

    standalone_columns: dict[str, dict[int, str]] = {
        "Name": {},
        "Type": {},
        "Features": {},
        "Dependencies": {},
        "Environment variables": {},
        "Scripts": {},
        "Description": {},
    }
    standalone_envs = (
        (env_name, config)
        for env_name, config in target_standalone_envs.items()
        if env_names or env_name not in matrix_envs
    )
    for i, (env_name, config) in enumerate(standalone_envs):
        if env_names and env_name not in env_names:
            continue

        environment = app.project.get_environment(env_name)

        standalone_columns["Name"][i] = env_name
        standalone_columns["Type"][i] = config["type"]

        if environment.features:
            standalone_columns["Features"][i] = "\n".join(environment.features)

        if environment.environment_dependencies_complex:
            standalone_columns["Dependencies"][i] = "\n".join(
                sorted({get_normalized_dependency(d) for d in environment.environment_dependencies_complex})
            )

        env_vars = dict(environment.env_vars)
        env_vars.pop(AppEnvVars.ENV_ACTIVE)
        if env_vars:
            standalone_columns["Environment variables"][i] = "\n".join(
                "=".join(item) for item in sorted(env_vars.items())
            )

        if environment.scripts:
            standalone_columns["Scripts"][i] = "\n".join(
                sorted(script for script in environment.scripts if app.verbose or not script.startswith("_"))
            )

        if environment.description:
            standalone_columns["Description"][i] = environment.description.strip()

    column_options = {}
    for title in matrix_columns:
        if title != "Description":
            column_options[title] = {"no_wrap": True}

    app.display_table(
        "" if hide_titles else "Standalone",
        standalone_columns,
        show_lines=True,
        column_options=column_options,
        force_ascii=force_ascii,
    )
    app.display_table(
        "" if hide_titles else "Matrices",
        matrix_columns,
        show_lines=True,
        column_options=column_options,
        force_ascii=force_ascii,
    )



from __future__ import annotations

import sys
from typing import Any

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def load_toml_data(data: str) -> dict[str, Any]:
    return tomllib.loads(data)


def load_toml_file(path: str) -> dict[str, Any]:
    with open(path, encoding="utf-8") as f:
        return tomllib.loads(f.read())



from __future__ import annotations

import json
import os

from hatchling.metadata.core import ProjectMetadata
from hatchling.metadata.utils import resolve_metadata_fields
from hatchling.plugin.manager import PluginManager

RUNNER: dict = {}


def main() -> None:
    project_root: str = RUNNER["project_root"]
    output_dir: str = RUNNER["output_dir"]

    project_metadata = ProjectMetadata(project_root, PluginManager())
    core_metadata = resolve_metadata_fields(project_metadata)
    for key, value in list(core_metadata.items()):
        if not value:
            core_metadata.pop(key)

    output = json.dumps(core_metadata)
    with open(os.path.join(output_dir, "output.json"), "w", encoding="utf-8") as f:
        f.write(output)


if __name__ == "__main__":
    main()



from hatchling.plugin.manager import PluginManager as _PluginManager


class PluginManager(_PluginManager):
    def initialize(self):
        super().initialize()

        from hatch.plugin import specs

        self.manager.add_hookspecs(specs)

    def hatch_register_environment(self):
        from hatch.env.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_environment_collector(self):
        from hatch.env.collectors.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_publisher(self):
        from hatch.publish.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_template(self):
        from hatch.template.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_locker(self):
        from hatch.env.lockers import hooks

        self.manager.register(hooks)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Locate environments")
@click.argument("env_name", default="default")
@click.pass_obj
def find(app: Application, env_name: str):
    """Locate environments."""
    app.ensure_environment_plugin_dependencies()

    environments = app.project.expand_environments(env_name)
    if not environments:
        app.abort(f"Environment `{env_name}` is not defined by project config")

    for env in environments:
        environment = app.project.get_environment(env)
        app.display(environment.find())



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hatch.python.distributions import DISTRIBUTIONS, ORDERED_DISTRIBUTIONS
from hatch.python.resolve import get_distribution
from hatch.utils.fs import temp_directory

if TYPE_CHECKING:
    from hatch.python.resolve import Distribution
    from hatch.utils.fs import Path


class InstalledDistribution:
    def __init__(self, path: Path, distribution: Distribution, metadata: dict[str, Any]) -> None:
        self.__path = path
        self.__current_dist = distribution
        self.__metadata = metadata

    @property
    def path(self) -> Path:
        return self.__path

    @property
    def name(self) -> str:
        return self.__current_dist.name

    @property
    def python_path(self) -> Path:
        return self.path / self.__current_dist.python_path

    @property
    def version(self) -> str:
        return self.__current_dist.version.base_version

    @property
    def metadata(self) -> dict[str, Any]:
        return self.__metadata

    def needs_update(self) -> bool:
        new_dist = get_distribution(self.__current_dist.name)
        return new_dist.version > self.__current_dist.version

    @classmethod
    def metadata_filename(cls) -> str:
        return "hatch-dist.json"


class PythonManager:
    def __init__(self, directory: Path) -> None:
        self.__directory = directory

    @property
    def directory(self) -> Path:
        return self.__directory

    def get_installed(self) -> dict[str, InstalledDistribution]:
        if not self.directory.is_dir():
            return {}

        import json

        installed_distributions: list[InstalledDistribution] = []
        for path in self.directory.iterdir():
            if not (path.name in DISTRIBUTIONS and path.is_dir()):
                continue

            metadata_file = path / InstalledDistribution.metadata_filename()
            if not metadata_file.is_file():
                continue

            metadata = json.loads(metadata_file.read_text())
            distribution = get_distribution(path.name, source=metadata.get("source", ""))
            if not (path / distribution.python_path).is_file():
                continue

            installed_distributions.append(InstalledDistribution(path, distribution, metadata))

        installed_distributions.sort(key=lambda d: ORDERED_DISTRIBUTIONS.index(d.name))
        return {dist.name: dist for dist in installed_distributions}

    def install(self, identifier: str) -> InstalledDistribution:
        import json

        from hatch.utils.network import download_file

        dist = get_distribution(identifier)
        path = self.directory / identifier
        self.directory.ensure_dir_exists()

        with temp_directory() as temp_dir:
            archive_path = temp_dir / dist.archive_name
            unpack_path = temp_dir / identifier
            download_file(archive_path, dist.source, follow_redirects=True)
            dist.unpack(archive_path, unpack_path)

            backup_path = path.with_suffix(".bak")
            if backup_path.is_dir():
                backup_path.wait_for_dir_removed()

            if path.is_dir():
                path.replace(backup_path)

            try:
                unpack_path.replace(path)
            except OSError:
                import shutil

                try:
                    shutil.move(str(unpack_path), str(path))
                except OSError:
                    path.wait_for_dir_removed()
                    if backup_path.is_dir():
                        backup_path.replace(path)

                    raise

        metadata = {"source": dist.source, "python_path": dist.python_path}
        metadata_file = path / InstalledDistribution.metadata_filename()
        metadata_file.write_text(json.dumps(metadata, indent=2))

        return InstalledDistribution(path, dist, metadata)

    @staticmethod
    def remove(dist: InstalledDistribution) -> None:
        dist.path.wait_for_dir_removed()



BUILD_BACKEND = "hatchling.build"
DEFAULT_BUILD_DIRECTORY = "dist"
DEFAULT_BUILD_SCRIPT = "hatch_build.py"
DEFAULT_CONFIG_FILE = "hatch.toml"


class BuildEnvVars:
    REQUESTED_TARGETS = "HATCH_BUILD_REQUESTED_TARGETS"
    LOCATION = "HATCH_BUILD_LOCATION"
    HOOKS_ONLY = "HATCH_BUILD_HOOKS_ONLY"
    NO_HOOKS = "HATCH_BUILD_NO_HOOKS"
    HOOKS_ENABLE = "HATCH_BUILD_HOOKS_ENABLE"
    HOOK_ENABLE_PREFIX = "HATCH_BUILD_HOOK_ENABLE_"
    CLEAN = "HATCH_BUILD_CLEAN"
    CLEAN_HOOKS_AFTER = "HATCH_BUILD_CLEAN_HOOKS_AFTER"



import os

from hatch.env.plugin.interface import EnvironmentInterface
from hatch.utils.env import PythonInfo


class SystemEnvironment(EnvironmentInterface):
    PLUGIN_NAME = "system"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.python_info = PythonInfo(self.platform)
        self.install_indicator = self.data_directory / str(self.root).encode("utf-8").hex()

    def find(self):
        return os.path.dirname(os.path.dirname(self.system_python))

    def create(self):
        self.install_indicator.touch()

    def remove(self):
        self.install_indicator.remove()

    def exists(self):
        return self.install_indicator.is_file()

    def install_project(self):
        self.platform.check_command(self.construct_pip_install_command([self.apply_features(str(self.root))]))

    def install_project_dev_mode(self):
        self.platform.check_command(
            self.construct_pip_install_command(["--editable", self.apply_features(str(self.root))])
        )

    def dependencies_in_sync(self):
        if not self.dependencies:
            return True

        from hatch.dep.sync import InstalledDistributions

        distributions = InstalledDistributions(
            sys_path=self.python_info.sys_path, environment=self.python_info.environment
        )
        return distributions.dependencies_in_sync(self.dependencies_complex)

    def sync_dependencies(self):
        self.platform.check_command(self.construct_pip_install_command(self.dependencies))



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from functools import cached_property
from textwrap import indent as indent_text
from typing import TYPE_CHECKING

import click
from rich.console import Console
from rich.style import Style
from rich.text import Text

if TYPE_CHECKING:
    from collections.abc import Callable

    from rich.status import Status


class TerminalStatus(ABC):
    @abstractmethod
    def stop(self) -> None: ...

    def __enter__(self) -> TerminalStatus:  # noqa: PYI034
        return self

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb): ...


class NullStatus(TerminalStatus):
    def stop(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class BorrowedStatus(TerminalStatus):
    def __init__(
        self,
        console: Console,
        *,
        is_interactive: bool,
        verbosity: int,
        spinner_style: str,
        waiting_style: Style | str,
        success_style: Style | str,
        initializer: Callable,
        finalizer: Callable,
    ):
        self.__console = console
        self.__is_interactive = is_interactive
        self.__verbosity = verbosity
        self.__spinner_style = spinner_style
        self.__waiting_style = waiting_style
        self.__success_style = success_style
        self.__initializer = initializer
        self.__finalizer = finalizer

        # This is the possibly active current status
        self.__status: Status | None = None

        # This is used as a stack to display the current message
        self.__messages: list[tuple[Text, str]] = []

    def stop(self) -> None:
        active = self.__active()
        if self.__status is not None:
            self.__status.stop()
            self.__finalizer()

        old_message, final_text = self.__messages[-1]
        if self.__verbosity > 0 and active:
            if not final_text:
                final_text = old_message.plain
                final_text = f"Finished {final_text[:1].lower()}{final_text[1:]}"

            self.__output(Text(final_text, style=self.__success_style))

    def __call__(self, message: str, final_text: str = "") -> BorrowedStatus:
        self.__messages.append((Text(message, style=self.__waiting_style), final_text))
        return self

    def __enter__(self) -> BorrowedStatus:  # noqa: PYI034
        if not self.__messages:
            return self

        message, _ = self.__messages[-1]
        if not self.__is_interactive:
            self.__output(message)
            return self

        if self.__status is None:
            self.__initializer()
        else:
            self.__status.stop()

        self.__status = self.__console.status(message, spinner=self.__spinner_style)
        self.__status.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        old_message, final_text = self.__messages.pop()
        if self.__verbosity > 0 and self.__active():
            if not final_text:
                final_text = old_message.plain
                final_text = f"Finished {final_text[:1].lower()}{final_text[1:]}"

            self.__output(Text(final_text, style=self.__success_style))

        if not self.__is_interactive:
            return

        self.__status.stop()
        if not self.__messages:
            self.__status = None
            self.__finalizer()
        else:
            self.__initializer()
            message, _ = self.__messages[-1]
            self.__status = self.__console.status(message, spinner=self.__spinner_style)
            self.__status.start()

    def __active(self) -> bool:
        return self.__status is not None and self.__status._live.is_started  # noqa: SLF001

    def __output(self, text):
        self.__console.stderr = True
        try:
            self.__console.print(text, overflow="ignore", no_wrap=True, crop=False)
        finally:
            self.__console.stderr = False


class Terminal:
    def __init__(self, *, verbosity: int, enable_color: bool | None, interactive: bool | None):
        # Force consistent output for test assertions
        self.testing = "HATCH_SELF_TESTING" in os.environ

        self.verbosity = verbosity
        self.console = Console(
            force_terminal=enable_color,
            force_interactive=interactive,
            no_color=enable_color is False,
            markup=False,
            emoji=False,
            highlight=False,
            legacy_windows=False if self.testing else None,
        )

        # Set defaults so we can pretty print before loading user config
        self._style_level_success: Style | str = "bold cyan"
        self._style_level_error: Style | str = "bold red"
        self._style_level_warning: Style | str = "bold yellow"
        self._style_level_waiting: Style | str = "bold magenta"
        # Default is simply bold rather than bold white for shells that have been configured with a white background
        self._style_level_info: Style | str = "bold"
        self._style_level_debug: Style | str = "bold"

        # Chosen as the default since it's compatible everywhere and looks nice
        self._style_spinner = "simpleDotsScrolling"

    @cached_property
    def kv_separator(self) -> Text:
        return self.style_warning("->")

    def style_success(self, text: str) -> Text:
        return Text(text, style=self._style_level_success)

    def style_error(self, text: str) -> Text:
        return Text(text, style=self._style_level_error)

    def style_warning(self, text: str) -> Text:
        return Text(text, style=self._style_level_warning)

    def style_waiting(self, text: str) -> Text:
        return Text(text, style=self._style_level_waiting)

    def style_info(self, text: str) -> Text:
        return Text(text, style=self._style_level_info)

    def style_debug(self, text: str) -> Text:
        return Text(text, style=self._style_level_debug)

    def initialize_styles(self, styles: dict):  # no cov
        from rich.errors import StyleSyntaxError
        from rich.spinner import Spinner

        # Lazily display errors so that they use the correct style
        errors = []

        for option, style in styles.items():
            attribute = f"_style_level_{option}"

            default_level = getattr(self, attribute, None)
            if default_level:
                try:
                    parsed_style = Style.parse(style)
                except StyleSyntaxError as e:  # no cov
                    errors.append(f"Invalid style definition for `{option}`, defaulting to `{default_level}`: {e}")
                    parsed_style = Style.parse(default_level)

                setattr(self, attribute, parsed_style)
            elif option == "spinner":
                try:
                    Spinner(style)
                except KeyError as e:
                    errors.append(
                        f"Invalid style definition for `{option}`, defaulting to `{self._style_spinner}`: {e.args[0]}"
                    )
                else:
                    self._style_spinner = style
            else:
                setattr(self, f"_style_{option}", style)

        return errors

    def display(self, text="", **kwargs):
        self.console.print(text, style=self._style_level_info, overflow="ignore", no_wrap=True, crop=False, **kwargs)

    def display_critical(self, text="", **kwargs):
        self.console.stderr = True
        try:
            self.console.print(
                text, style=self._style_level_error, overflow="ignore", no_wrap=True, crop=False, **kwargs
            )
        finally:
            self.console.stderr = False

    def display_error(self, text="", *, stderr=True, indent=None, link=None, **kwargs):
        if self.verbosity < -2:  # noqa: PLR2004
            return

        self._output(text, self._style_level_error, stderr=stderr, indent=indent, link=link, **kwargs)

    def display_warning(self, text="", *, stderr=True, indent=None, link=None, **kwargs):
        if self.verbosity < -1:
            return

        self._output(text, self._style_level_warning, stderr=stderr, indent=indent, link=link, **kwargs)

    def display_info(self, text="", *, stderr=True, indent=None, link=None, **kwargs):
        if self.verbosity < 0:
            return

        self._output(text, self._style_level_info, stderr=stderr, indent=indent, link=link, **kwargs)

    def display_success(self, text="", *, stderr=True, indent=None, link=None, **kwargs):
        if self.verbosity < 0:
            return

        self._output(text, self._style_level_success, stderr=stderr, indent=indent, link=link, **kwargs)

    def display_waiting(self, text="", *, stderr=True, indent=None, link=None, **kwargs):
        if self.verbosity < 0:
            return

        self._output(text, self._style_level_waiting, stderr=stderr, indent=indent, link=link, **kwargs)

    def display_debug(self, text="", level=1, *, stderr=True, indent=None, link=None, **kwargs):
        if not 1 <= level <= 3:  # noqa: PLR2004
            error_message = "Debug output can only have verbosity levels between 1 and 3 (inclusive)"
            raise ValueError(error_message)

        if self.verbosity < level:
            return

        self._output(text, self._style_level_debug, stderr=stderr, indent=indent, link=link, **kwargs)

    def display_mini_header(self, text, *, stderr=False, indent=None, link=None):
        if self.verbosity < 0:
            return

        self.display_info("[", stderr=stderr, indent=indent, end="")
        self.display_success(text, stderr=stderr, link=link, end="")
        self.display_info("]", stderr=stderr)

    def display_header(self, title=""):
        self.console.rule(Text(title, self._style_level_success))

    def display_syntax(self, *args, **kwargs):
        from rich.syntax import Syntax

        kwargs.setdefault("background_color", "default" if self.testing else None)
        self.output(Syntax(*args, **kwargs))

    def display_markdown(self, text, **kwargs):  # no cov
        from rich.markdown import Markdown

        self.output(Markdown(text), **kwargs)

    def display_pair(self, key, value):
        self.output(self.style_success(key), self.kv_separator, value)

    def display_table(self, title, columns, *, show_lines=False, column_options=None, force_ascii=False, num_rows=0):
        from rich.table import Table

        if column_options is None:
            column_options = {}

        table_options = {}
        if force_ascii:
            from rich.box import ASCII_DOUBLE_HEAD

            table_options["box"] = ASCII_DOUBLE_HEAD
            table_options["safe_box"] = True

        table = Table(title=title, show_lines=show_lines, title_style="", **table_options)
        columns = dict(columns)

        for column_title, indices in list(columns.items()):
            if indices:
                table.add_column(column_title, style="bold", **column_options.get(column_title, {}))
            else:
                columns.pop(column_title)

        if not columns:
            return

        for i in range(num_rows or max(map(max, columns.values())) + 1):
            row = [indices.get(i, "") for indices in columns.values()]
            if any(row):
                table.add_row(*row)

        self.output(table)

    @cached_property
    def status(self) -> BorrowedStatus:
        return BorrowedStatus(
            self.console,
            is_interactive=self.console.is_interactive,
            verbosity=self.verbosity,
            spinner_style=self._style_spinner,
            waiting_style=self._style_level_waiting,
            success_style=self._style_level_success,
            initializer=lambda: setattr(self.platform, "displaying_status", True),  # type: ignore[attr-defined]
            finalizer=lambda: setattr(self.platform, "displaying_status", False),  # type: ignore[attr-defined]
        )

    def status_if(self, *args, condition: bool, **kwargs) -> TerminalStatus:
        return self.status(*args, **kwargs) if condition else NullStatus()

    def _output(self, text="", style=None, *, stderr=False, indent=None, link=None, **kwargs):
        if indent:
            text = indent_text(text, indent)

        if link:
            style = style.update_link(self.platform.format_file_uri(link))

        self.output(text, stderr=stderr, style=style, **kwargs)

    def output(self, *args, stderr=False, **kwargs):
        kwargs.setdefault("overflow", "ignore")
        kwargs.setdefault("no_wrap", True)
        kwargs.setdefault("crop", False)

        if not stderr:
            self.console.print(*args, **kwargs)
        else:
            self.console.stderr = True
            try:
                self.console.print(*args, **kwargs)
            finally:
                self.console.stderr = False

    @staticmethod
    def prompt(text, **kwargs):
        return click.prompt(text, **kwargs)

    @staticmethod
    def confirm(text, **kwargs):
        return click.confirm(text, **kwargs)



import os

import click


@click.group(short_help="Manage the config file")
def config():
    pass


@config.command(short_help="Open the config location in your file manager")
@click.pass_obj
def explore(app):
    """Open the config location in your file manager."""
    click.launch(str(app.config_file.path), locate=True)


@config.command(short_help="Show the location of the config file")
@click.pass_obj
def find(app):
    """Show the location of the config file."""
    app.display(str(app.config_file.path))


@config.command(short_help="Show the contents of the config file")
@click.option("--all", "-a", "all_keys", is_flag=True, help="Do not scrub secret fields")
@click.pass_obj
def show(app, all_keys):
    """Show the contents of the config file."""
    if not app.config_file.path.is_file():  # no cov
        app.display_critical("No config file found! Please try `hatch config restore`.")
    else:
        text = app.config_file.read() if all_keys else app.config_file.read_scrubbed()
        app.display_syntax(text.rstrip(), "toml")


@config.command(short_help="Update the config file with any new fields")
@click.pass_obj
def update(app):  # no cov
    """Update the config file with any new fields."""
    app.config_file.update()
    app.display_success("Settings were successfully updated.")


@config.command(short_help="Restore the config file to default settings")
@click.pass_obj
def restore(app):
    """Restore the config file to default settings."""
    app.config_file.restore()
    app.display_success("Settings were successfully restored.")


@config.command("set", short_help="Assign values to config file entries")
@click.argument("key")
@click.argument("value", required=False)
@click.pass_obj
def set_value(app, key, value):
    """
    Assign values to config file entries. If the value is omitted,
    you will be prompted, with the input hidden if it is sensitive.
    """
    from fnmatch import fnmatch

    import tomlkit

    from hatch.config.model import ConfigurationError, RootConfig
    from hatch.config.utils import create_toml_document, save_toml_document

    scrubbing = key.startswith("publish.")
    if value is None:
        value = click.prompt(f"Value for `{key}`", hide_input=scrubbing)

    setting_project_location = bool(fnmatch(key, "projects.*") or fnmatch(key, "projects.*.location"))
    if setting_project_location and not value.startswith("~"):
        value = os.path.abspath(value)

    user_config = new_config = tomlkit.parse(app.config_file.read())

    data = [value]
    data.extend(reversed(key.split(".")))
    key = data.pop()
    value = data.pop()

    # Use a separate mapping to show only what has changed in the end
    branch_config_root = branch_config = {}

    # Consider dots as keys
    while data:
        default_branch = {value: ""}
        branch_config[key] = default_branch
        branch_config = branch_config[key]

        new_value = new_config.get(key)
        if not hasattr(new_value, "get"):
            new_value = default_branch

        new_config[key] = new_value
        new_config = new_config[key]

        key = value
        value = data.pop()

    if value.startswith(("{", "[")):
        from ast import literal_eval

        value = literal_eval(value)
    elif value.lower() == "true":
        value = True
    elif value.lower() == "false":
        value = False

    branch_config[key] = new_config[key] = value

    # https://github.com/sdispater/tomlkit/issues/48
    if new_config.__class__.__name__ == "Table":  # no cov
        table_body = getattr(new_config.value, "body", [])
        possible_whitespace = table_body[-2:]
        if len(possible_whitespace) == 2:  # noqa: PLR2004
            for key, item in possible_whitespace:
                if key is not None:
                    break
                if item.__class__.__name__ != "Whitespace":
                    break
            else:
                del table_body[-2]

    try:
        RootConfig(user_config).parse_fields()
    except ConfigurationError as e:
        app.display_error(str(e))
        app.abort()
    else:
        if not user_config["project"] and setting_project_location:
            project = next(iter(branch_config_root["projects"]))
            user_config["project"] = project
            branch_config_root["project"] = project

        save_toml_document(user_config, app.config_file.path)

    document = create_toml_document(branch_config_root)
    if scrubbing and "publish" in document:
        for data in document["publish"].values():
            for field in list(data):
                data[field] = "<...>"

    from rich.syntax import Syntax

    app.display_success("New setting:")
    app.output(Syntax(tomlkit.dumps(document).rstrip(), "toml", background_color="default"))



from hatch.template import File
from hatch.utils.fs import Path


class PackageEntryPoint(File):
    TEMPLATE = """\
import sys

if __name__ == "__main__":
    from {package_name}.cli import {package_name}

    sys.exit({package_name}())
"""

    def __init__(
        self,
        template_config: dict,
        plugin_config: dict,  # noqa: ARG002
    ):
        super().__init__(Path(template_config["package_name"], "__main__.py"), self.TEMPLATE.format(**template_config))


class CommandLinePackage(File):
    TEMPLATE = """\
import click

from {package_name}.__about__ import __version__


@click.group(context_settings={{"help_option_names": ["-h", "--help"]}}, invoke_without_command=True)
@click.version_option(version=__version__, prog_name="{project_name}")
def {package_name}():
    click.echo("Hello world!")
"""

    def __init__(
        self,
        template_config: dict,
        plugin_config: dict,  # noqa: ARG002
    ):
        super().__init__(
            Path(template_config["package_name"], "cli", "__init__.py"), self.TEMPLATE.format(**template_config)
        )



from __future__ import annotations

from functools import cached_property
from os import environ
from typing import TYPE_CHECKING, Any

from hatch.utils.platform import get_platform_name

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.utils.fs import Path

RESERVED_OPTIONS = {
    "builder": bool,
    "dependencies": list,
    "dependency-groups": list,
    "extra-dependencies": list,
    "dev-mode": bool,
    "env-exclude": list,
    "env-include": list,
    "env-vars": dict,
    "features": list,
    "lock-filename": str,
    "locker": str,
    "locked": bool,
    "matrix-name-format": str,
    "platforms": list,
    "post-install-commands": list,
    "pre-install-commands": list,
    "python": str,
    "scripts": dict,
    "skip-install": bool,
    "type": str,
    "workspace": dict,
}


def apply_overrides(env_name, source, condition, condition_value, options, new_config, option_types=None):
    if option_types is None:
        option_types = RESERVED_OPTIONS

    for raw_option, data in options.items():
        _, separator, option = raw_option.rpartition("set-")
        overwrite = bool(separator)

        # Prevent manipulation of reserved options
        if option_types is not RESERVED_OPTIONS and option in RESERVED_OPTIONS:
            continue

        override_type = option_types.get(option)
        if option == "workspace":
            _apply_override_to_workspace(
                env_name, option, data, source, condition, condition_value, new_config, overwrite
            )
        elif override_type in TYPE_OVERRIDES:
            TYPE_OVERRIDES[override_type](
                env_name, option, data, source, condition, condition_value, new_config, overwrite
            )
        elif isinstance(data, dict) and "value" in data:
            if _resolve_condition(env_name, option, source, condition, condition_value, data):
                new_config[option] = data["value"]
        elif option_types is not RESERVED_OPTIONS:
            message = (
                f"Untyped option `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                f"must be defined as a table with a `value` key"
            )
            raise ValueError(message)


def _apply_override_to_mapping(env_name, option, data, source, condition, condition_value, new_config, overwrite):
    new_mapping = {}
    if isinstance(data, str):
        key, separator, value = data.partition("=")
        if not separator:
            value = condition_value
        new_mapping[key] = value
    elif isinstance(data, list):
        for i, entry in enumerate(data, 1):
            if isinstance(entry, str):
                key, separator, value = entry.partition("=")
                if not separator:
                    value = condition_value
                new_mapping[key] = value
            elif isinstance(entry, dict):
                if "key" not in entry:
                    message = (
                        f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                        f"must have an option named `key`"
                    )
                    raise ValueError(message)
                key = entry["key"]
                if not isinstance(key, str):
                    message = (
                        f"Option `key` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                        f"{condition}.{option}` must be a string"
                    )
                    raise TypeError(message)

                if not key:
                    message = (
                        f"Option `key` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                        f"{condition}.{option}` cannot be an empty string"
                    )
                    raise ValueError(message)

                value = entry.get("value", condition_value)
                if not isinstance(value, str):
                    message = (
                        f"Option `value` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                        f"{condition}.{option}` must be a string"
                    )
                    raise TypeError(message)
                if _resolve_condition(env_name, option, source, condition, condition_value, entry, i):
                    new_mapping[key] = value
            else:
                message = (
                    f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                    f"must be a string or an inline table"
                )
                raise TypeError(message)
    else:
        message = (
            f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` must be a string or an array"
        )
        raise TypeError(message)

    if overwrite:
        new_config[option] = new_mapping
    elif option in new_config:
        new_config[option].update(new_mapping)
    elif new_mapping:
        new_config[option] = new_mapping


def _apply_override_to_array(env_name, option, data, source, condition, condition_value, new_config, overwrite):
    if not isinstance(data, list):
        message = f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` must be an array"
        raise TypeError(message)

    new_array = []
    for i, entry in enumerate(data, 1):
        if isinstance(entry, str):
            new_array.append(entry)
        elif isinstance(entry, dict):
            if "value" not in entry:
                message = (
                    f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                    f"must have an option named `value`"
                )
                raise ValueError(message)
            value = entry["value"]
            if not isinstance(value, str):
                message = (
                    f"Option `value` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                    f"{condition}.{option}` must be a string"
                )
                raise TypeError(message)

            if not value:
                message = (
                    f"Option `value` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                    f"{condition}.{option}` cannot be an empty string"
                )
                raise ValueError(message)
            if _resolve_condition(env_name, option, source, condition, condition_value, entry, i):
                new_array.append(value)
        else:
            message = (
                f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                f"must be a string or an inline table"
            )
            raise TypeError(message)

    if overwrite:
        new_config[option] = new_array
    elif option in new_config:
        new_config[option].extend(new_array)
    elif new_array:
        new_config[option] = new_array


def _apply_override_to_string(
    env_name,
    option,
    data,
    source,
    condition,
    condition_value,
    new_config,
    overwrite,  # noqa: ARG001
):
    if isinstance(data, str):
        new_config[option] = data
    elif isinstance(data, dict):
        if "value" not in data:
            message = (
                f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                f"must have an option named `value`"
            )
            raise ValueError(message)
        value = data["value"]
        if not isinstance(value, str):
            message = (
                f"Option `value` in field `tool.hatch.envs.{env_name}.overrides.{source}."
                f"{condition}.{option}` must be a string"
            )
            raise TypeError(message)
        if _resolve_condition(env_name, option, source, condition, condition_value, data):
            new_config[option] = value
    elif isinstance(data, list):
        for i, entry in enumerate(data, 1):
            if isinstance(entry, str):
                new_config[option] = entry
                break

            if isinstance(entry, dict):
                if "value" not in entry:
                    message = (
                        f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                        f"must have an option named `value`"
                    )
                    raise ValueError(message)
                value = entry["value"]
                if not isinstance(value, str):
                    message = (
                        f"Option `value` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                        f"{condition}.{option}` must be a string"
                    )
                    raise TypeError(message)
                if _resolve_condition(env_name, option, source, condition, condition_value, entry, i):
                    new_config[option] = value
                    break
            else:
                message = (
                    f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                    f"must be a string or an inline table"
                )
                raise TypeError(message)
    else:
        message = (
            f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
            f"must be a string, inline table, or an array"
        )
        raise TypeError(message)


def _apply_override_to_boolean(
    env_name,
    option,
    data,
    source,
    condition,
    condition_value,
    new_config,
    overwrite,  # noqa: ARG001
):
    if isinstance(data, bool):
        new_config[option] = data
    elif isinstance(data, dict):
        if "value" not in data:
            message = (
                f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                f"must have an option named `value`"
            )
            raise ValueError(message)
        value = data["value"]
        if not isinstance(value, bool):
            message = (
                f"Option `value` in field `tool.hatch.envs.{env_name}.overrides.{source}."
                f"{condition}.{option}` must be a boolean"
            )
            raise TypeError(message)
        if _resolve_condition(env_name, option, source, condition, condition_value, data):
            new_config[option] = value
    elif isinstance(data, list):
        for i, entry in enumerate(data, 1):
            if isinstance(entry, bool):
                new_config[option] = entry
                break

            if isinstance(entry, dict):
                if "value" not in entry:
                    message = (
                        f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                        f"must have an option named `value`"
                    )
                    raise ValueError(message)
                value = entry["value"]
                if not isinstance(value, bool):
                    message = (
                        f"Option `value` in entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}."
                        f"{condition}.{option}` must be a boolean"
                    )
                    raise TypeError(message)
                if _resolve_condition(env_name, option, source, condition, condition_value, entry, i):
                    new_config[option] = value
                    break
            else:
                message = (
                    f"Entry #{i} in field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
                    f"must be a boolean or an inline table"
                )
                raise TypeError(message)
    else:
        message = (
            f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` "
            f"must be a boolean, inline table, or an array"
        )
        raise TypeError(message)


def _apply_override_to_workspace(env_name, option, data, source, condition, condition_value, new_config, overwrite):
    """Handle workspace dict with nested members/exclude/parallel."""
    if not isinstance(data, dict):
        message = f"Field `tool.hatch.envs.{env_name}.overrides.{source}.{condition}.{option}` must be a table"
        raise TypeError(message)

    # Get or create workspace dict
    workspace = {} if overwrite else new_config.setdefault(option, {})

    for key, value in data.items():
        if key in {"members", "exclude"}:
            # Delegate to array handler - pass workspace dict
            _apply_override_to_array(env_name, key, value, source, condition, condition_value, workspace, overwrite)
        elif key == "parallel":
            # Delegate to boolean handler - pass workspace dict
            _apply_override_to_boolean(env_name, key, value, source, condition, condition_value, workspace, overwrite)
        else:
            message = f"Unknown workspace option: {key}"
            raise ValueError(message)

    # Update new_config with the workspace dict
    if overwrite or workspace:
        new_config[option] = workspace


def _resolve_condition(env_name, option, source, condition, condition_value, condition_config, condition_index=None):
    location = "field" if condition_index is None else f"entry #{condition_index} in field"

    if "if" in condition_config:
        allowed_values = condition_config["if"]
        if not isinstance(allowed_values, list):
            message = (
                f"Option `if` in {location} `tool.hatch.envs.{env_name}.overrides.{source}."
                f"{condition}.{option}` must be an array"
            )
            raise TypeError(message)

        if condition_value not in allowed_values:
            return False

    if "platform" in condition_config:
        allowed_platforms = condition_config["platform"]
        if not isinstance(allowed_platforms, list):
            message = (
                f"Option `platform` in {location} `tool.hatch.envs.{env_name}.overrides.{source}."
                f"{condition}.{option}` must be an array"
            )
            raise TypeError(message)

        for i, entry in enumerate(allowed_platforms, 1):
            if not isinstance(entry, str):
                message = (
                    f"Item #{i} in option `platform` in {location} `tool.hatch.envs.{env_name}.overrides.{source}."
                    f"{condition}.{option}` must be a string"
                )
                raise TypeError(message)

        if get_platform_name() not in allowed_platforms:
            return False

    if "env" in condition_config:
        env_vars = condition_config["env"]
        if not isinstance(env_vars, list):
            message = (
                f"Option `env` in {location} `tool.hatch.envs.{env_name}.overrides.{source}."
                f"{condition}.{option}` must be an array"
            )
            raise TypeError(message)

        required_env_vars = {}
        for i, entry in enumerate(env_vars, 1):
            if not isinstance(entry, str):
                message = (
                    f"Item #{i} in option `env` in {location} `tool.hatch.envs.{env_name}.overrides.{source}."
                    f"{condition}.{option}` must be a string"
                )
                raise TypeError(message)

            # Allow matching empty strings
            if "=" in entry:
                env_var, _, value = entry.partition("=")
                required_env_vars[env_var] = value
            else:
                required_env_vars[entry] = None

        for env_var, value in required_env_vars.items():
            if env_var not in environ or (value is not None and value != environ[env_var]):
                return False

    return True


TYPE_OVERRIDES = {
    dict: _apply_override_to_mapping,
    list: _apply_override_to_array,
    str: _apply_override_to_string,
    bool: _apply_override_to_boolean,
}


class EnvironmentMetadata:
    def __init__(self, data_dir: Path, project_path: Path):
        self.__data_dir = data_dir
        self.__project_path = project_path

    def dependency_hash(self, environment: EnvironmentInterface) -> str:
        return self._read(environment).get("dependency_hash", "")

    def update_dependency_hash(self, environment: EnvironmentInterface, dependency_hash: str) -> None:
        metadata = self._read(environment)
        metadata["dependency_hash"] = dependency_hash
        self._write(environment, metadata)

    def reset(self, environment: EnvironmentInterface) -> None:
        self._metadata_file(environment).unlink(missing_ok=True)

    def _read(self, environment: EnvironmentInterface) -> dict[str, Any]:
        import json

        metadata_file = self._metadata_file(environment)
        if not metadata_file.is_file():
            return {}

        return json.loads(metadata_file.read_text())

    def _write(self, environment: EnvironmentInterface, metadata: dict[str, Any]) -> None:
        import json

        metadata_file = self._metadata_file(environment)
        metadata_file.parent.ensure_dir_exists()
        metadata_file.write_text(json.dumps(metadata))

    def _metadata_file(self, environment: EnvironmentInterface) -> Path:
        from hatch.env.internal import is_isolated_environment

        if is_isolated_environment(environment.name, environment.config):
            return self.__data_dir / ".internal" / f"{environment.name}.json"

        return self._storage_dir / environment.config["type"] / f"{environment.name}.json"

    @cached_property
    def _storage_dir(self) -> Path:
        return self.__data_dir / self.__project_path.id



import click

from hatch.cli.env.create import create
from hatch.cli.env.find import find
from hatch.cli.env.lock import lock
from hatch.cli.env.prune import prune
from hatch.cli.env.remove import remove
from hatch.cli.env.run import run
from hatch.cli.env.show import show


@click.group(short_help="Manage project environments")
def env():
    pass


env.add_command(create)
env.add_command(find)
env.add_command(lock)
env.add_command(prune)
env.add_command(remove)
env.add_command(run)
env.add_command(show)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from hatch.utils.fs import Path

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from types import FrameType

    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.utils.platform import Platform


def detect_shell(platform: Platform) -> tuple[str, str]:
    import shellingham

    try:
        return shellingham.detect_shell()
    except shellingham.ShellDetectionFailure:
        path = platform.default_shell
        return Path(path).stem, path


class ShellManager:
    def __init__(self, environment: EnvironmentInterface) -> None:
        self.environment = environment

    def enter_cmd(self, path: str, args: Iterable[str], exe_dir: Path) -> None:  # noqa: ARG002
        self.environment.platform.exit_with_command([path or "cmd", "/k", str(exe_dir / "activate.bat")])

    def enter_powershell(self, path: str, args: Iterable[str], exe_dir: Path) -> None:  # noqa: ARG002
        self.environment.platform.exit_with_command([
            path or "powershell",
            "-executionpolicy",
            "bypass",
            "-NoExit",
            "-NoLogo",
            "-File",
            str(exe_dir / "activate.ps1"),
        ])

    def enter_pwsh(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        self.enter_powershell(path or "pwsh", args, exe_dir)

    def enter_xonsh(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        if self.environment.platform.windows:
            with self.environment:
                self.environment.platform.exit_with_command([
                    path or "xonsh",
                    *(args or ["-i"]),
                    "-D",
                    f"VIRTUAL_ENV={exe_dir.parent.name}",
                ])
        else:
            self.spawn_linux_shell(
                path or "xonsh",
                [*(args or ["-i"]), "-D", f"VIRTUAL_ENV={exe_dir.parent.name}"],
                # Just in case pyenv works with xonsh, supersede it.
                callback=lambda terminal: terminal.sendline(f"$PATH.insert(0, {str(exe_dir)!r})"),
            )

    def enter_bash(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        if self.environment.platform.windows:
            self.environment.platform.exit_with_command([
                path or "bash",
                "--init-file",
                exe_dir / "activate",
                *(args or ["-i"]),
            ])
        else:
            self.spawn_linux_shell(path or "bash", args or ["-i"], script=exe_dir / "activate")

    def enter_fish(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        self.spawn_linux_shell(path or "fish", args or ["-i"], script=exe_dir / "activate.fish")

    def enter_zsh(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        self.spawn_linux_shell(path or "zsh", args or ["-i"], script=exe_dir / "activate")

    def enter_ash(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        self.spawn_linux_shell(path or "ash", args or ["-i"], script=exe_dir / "activate")

    def enter_nu(self, path: str, args: Iterable[str], exe_dir: Path) -> None:  # noqa: ARG002
        executable = path or "nu"
        activation_script = exe_dir / "activate.nu"
        self.environment.platform.exit_with_command([executable, "-e", f"overlay use {str(activation_script)!r}"])

    def enter_tcsh(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        self.spawn_linux_shell(path or "tcsh", args or ["-i"], script=exe_dir / "activate.csh")

    def enter_csh(self, path: str, args: Iterable[str], exe_dir: Path) -> None:
        self.spawn_linux_shell(path or "csh", args or ["-i"], script=exe_dir / "activate.csh")

    if sys.platform == "win32":

        def spawn_linux_shell(
            self,
            path: str,
            args: Iterable[str] | None = None,
            *,
            script: Path | None = None,
            callback: Callable | None = None,
        ) -> None:
            raise NotImplementedError

    else:

        def spawn_linux_shell(
            self,
            path: str,
            args: Iterable[str] | None = None,
            *,
            script: Path | None = None,
            callback: Callable | None = None,
        ) -> None:
            import shutil
            import signal

            import pexpect

            columns, lines = shutil.get_terminal_size()
            # pexpect only accepts lists
            terminal = pexpect.spawn(path, args=list(args or ()), dimensions=(lines, columns))

            def sigwinch_passthrough(sig: int, data: FrameType | None) -> None:  # noqa: ARG001
                new_columns, new_lines = shutil.get_terminal_size()
                terminal.setwinsize(new_lines, new_columns)

            signal.signal(signal.SIGWINCH, sigwinch_passthrough)

            if script is not None:
                terminal.sendline(f'source "{script}"')

            if callback is not None:
                callback(terminal)

            terminal.interact(escape_character=None)
            terminal.close()

            self.environment.platform.exit_with_code(terminal.exitstatus)



from __future__ import annotations

import os
import sys
from functools import cached_property
from typing import TYPE_CHECKING, cast

from hatch.cli.terminal import Terminal
from hatch.config.user import ConfigFile, RootConfig
from hatch.project.core import Project
from hatch.utils.fs import Path
from hatch.utils.platform import Platform
from hatch.utils.runner import ExecutionContext

if TYPE_CHECKING:
    from collections.abc import Generator

    from hatch.dep.core import Dependency
    from hatch.env.plugin.interface import EnvironmentInterface


class Application(Terminal):
    def __init__(self, exit_func, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.platform = Platform(self.output)
        self.__exit_func = exit_func

        self.config_file = ConfigFile()
        self.quiet = self.verbosity < 0
        self.verbose = self.verbosity > 0

        # Lazily set these as we acquire more knowledge about the environment
        self.data_dir = cast(Path, None)
        self.cache_dir = cast(Path, None)
        self.project = cast(Project, None)
        self.env = cast(str, None)
        self.env_active = cast(str, None)

    @property
    def plugins(self):
        return self.project.plugin_manager

    @property
    def config(self) -> RootConfig:
        return self.config_file.model

    def get_environment(self, env_name: str | None = None) -> EnvironmentInterface:
        return self.project.get_environment(env_name)

    def prepare_environment(self, environment: EnvironmentInterface, *, keep_env: bool = False):
        self.project.prepare_environment(environment, keep_env=keep_env)

    def run_shell_commands(self, context: ExecutionContext) -> None:
        with context.env.command_context():
            try:
                resolved_commands = list(context.env.resolve_commands(context.shell_commands))
            except Exception as e:  # noqa: BLE001
                self.abort(str(e))

            first_error_code = None
            should_display_command = not context.hide_commands and (self.verbose or len(resolved_commands) > 1)
            for i, raw_command in enumerate(resolved_commands, 1):
                if should_display_command:
                    self.display_info(f"{context.source} [{i}] | {raw_command}")

                command = raw_command
                continue_on_error = context.force_continue
                if raw_command.startswith("- "):
                    continue_on_error = True
                    command = command[2:]

                process = context.env.run_shell_command(command)
                sys.stdout.flush()
                sys.stderr.flush()
                if process.returncode:
                    first_error_code = first_error_code or process.returncode
                    if continue_on_error:
                        continue

                    if context.show_code_on_error:
                        self.abort(f"Failed with exit code: {process.returncode}", code=process.returncode)
                    else:
                        self.abort(code=process.returncode)

            if first_error_code and context.force_continue:
                self.abort(code=first_error_code)

    def runner_context(
        self,
        environments: list[str],
        *,
        ignore_compat: bool = False,
        display_header: bool = False,
        keep_env: bool = False,
    ) -> Generator[ExecutionContext, None, None]:
        if self.verbose or len(environments) > 1:
            display_header = True

        any_compatible = False
        incompatible = {}
        with self.project.ensure_cwd():
            for env_name in environments:
                environment = self.get_environment(env_name)
                if not environment.exists():
                    try:
                        environment.check_compatibility()
                    except Exception as e:  # noqa: BLE001
                        if ignore_compat:
                            incompatible[environment.name] = str(e)
                            continue

                        self.abort(f"Environment `{env_name}` is incompatible: {e}")

                any_compatible = True
                if display_header:
                    self.display_header(environment.name)

                context = ExecutionContext(environment)
                yield context

                self.prepare_environment(environment, keep_env=keep_env)
                self.execute_context(context)

        if incompatible:
            num_incompatible = len(incompatible)
            padding = "\n" if any_compatible else ""
            self.display_warning(
                f"{padding}Skipped {num_incompatible} incompatible environment{'s' if num_incompatible > 1 else ''}:"
            )
            for env_name, reason in incompatible.items():
                self.display_warning(f"{env_name} -> {reason}")

    def execute_context(self, context: ExecutionContext) -> None:
        from hatch.utils.structures import EnvVars

        with EnvVars(context.env_vars):
            self.run_shell_commands(context)

    def ensure_environment_plugin_dependencies(self) -> None:
        self.ensure_plugin_dependencies(
            self.project.config.env_requires_complex, wait_message="Syncing environment plugin requirements"
        )

    def ensure_plugin_dependencies(self, dependencies: list[Dependency], *, wait_message: str) -> None:
        if not dependencies:
            return

        from hatch.dep.sync import InstalledDistributions
        from hatch.env.utils import add_verbosity_flag

        if app_path := os.environ.get("PYAPP"):
            from hatch.utils.env import PythonInfo

            management_command = os.environ["PYAPP_COMMAND_NAME"]
            executable = self.platform.check_command_output([app_path, management_command, "python-path"]).strip()
            python_info = PythonInfo(self.platform, executable=executable)
            distributions = InstalledDistributions(sys_path=python_info.sys_path)
            if distributions.dependencies_in_sync(dependencies):
                return

            pip_command = [app_path, management_command, "pip"]
        else:
            distributions = InstalledDistributions()
            if distributions.dependencies_in_sync(dependencies):
                return

            pip_command = [sys.executable, "-u", "-m", "pip"]

        pip_command.extend(["install", "--disable-pip-version-check"])

        # Default to -1 verbosity
        add_verbosity_flag(pip_command, self.verbosity, adjustment=-1)

        pip_command.extend(str(dependency) for dependency in dependencies)

        with self.status(wait_message):
            self.platform.check_command(pip_command)

    def get_env_directory(self, environment_type):
        directories = self.config.dirs.env

        if environment_type in directories:
            path = Path(directories[environment_type]).expand()
            if os.path.isabs(path):
                return path

            return self.project.location / path

        return self.data_dir / "env" / environment_type

    def get_python_manager(self, directory: str | None = None):
        from hatch.python.core import PythonManager

        configured_dir = directory or self.config.dirs.python
        if configured_dir == "isolated":
            return PythonManager(self.data_dir / "pythons")

        return PythonManager(Path(configured_dir).expand())

    @cached_property
    def shell_data(self) -> tuple[str, str]:
        from hatch.utils.shells import detect_shell

        return detect_shell(self.platform)

    def abort(self, text="", code=1, **kwargs):
        if text:
            self.display_error(text, **kwargs)
        self.__exit_func(code)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterable


def parse_script_command(command: str) -> tuple[str, str, bool]:
    possible_script, _, args = command.partition(" ")
    if possible_script == "-":
        ignore_exit_code = True
        possible_script, _, args = args.partition(" ")
    else:
        ignore_exit_code = False

    return possible_script, args, ignore_exit_code


def format_script_commands(*, commands: list[str], args: str, ignore_exit_code: bool) -> Iterable[str]:
    for command in commands:
        if args:
            yield f"{command} {args}"
        elif ignore_exit_code and not command.startswith("- "):
            yield f"- {command}"
        else:
            yield command


def parse_inline_script_metadata(script: str) -> dict[str, Any] | None:
    """
    https://peps.python.org/pep-0723/#reference-implementation
    """
    import re

    from hatch.utils.toml import load_toml_data

    block_type = "script"
    pattern = re.compile(r"(?m)^# /// (?P<type>[a-zA-Z0-9-]+)$\s(?P<content>(^#(| .*)$\s)+)^# ///$")
    matches = list(filter(lambda m: m.group("type") == block_type, pattern.finditer(script)))
    if len(matches) > 1:
        message = f"Multiple inline metadata blocks found for type: {block_type}"
        raise ValueError(message)

    if len(matches) == 1:
        content = "".join(
            line[2:] if line.startswith("# ") else line[1:]
            for line in matches[0].group("content").splitlines(keepends=True)
        )
        return load_toml_data(content)

    return None



from __future__ import annotations

import click

from hatch.cli.check.code import code
from hatch.cli.check.fmt import fmt
from hatch.cli.check.types import types


@click.group(short_help="Check source code", invoke_without_command=True)
@click.option("--fix", is_flag=True, help="Fix issues rather than just reporting them")
@click.pass_context
def check(ctx: click.Context, *, fix: bool):
    """Check source code for issues (linting, formatting, type checking).

    When invoked without a subcommand, runs all checks (code, fmt, types).
    """
    if ctx.invoked_subcommand is not None:
        return

    ctx.invoke(code, fix=fix)
    ctx.invoke(fmt, fix=fix)
    ctx.invoke(types)


check.add_command(code)
check.add_command(fmt)
check.add_command(types)



from hatch.env.collectors.plugin.interface import EnvironmentCollectorInterface
from hatch.env.internal import get_internal_env_config
from hatch.env.utils import ensure_valid_environment


class DefaultEnvironmentCollector(EnvironmentCollectorInterface):
    PLUGIN_NAME = "default"

    def get_initial_config(self):  # noqa: PLR6301
        default_config = {}
        ensure_valid_environment(default_config)

        return {"default": default_config, **get_internal_env_config()}



from __future__ import annotations

import os
from typing import Any

from hatch.env.collectors.plugin.interface import EnvironmentCollectorInterface
from hatch.plugin.constants import DEFAULT_CUSTOM_SCRIPT
from hatch.plugin.utils import load_plugin_from_script


class CustomEnvironmentCollector:
    PLUGIN_NAME = "custom"

    def __new__(  # type: ignore[misc]
        cls,
        root: str,
        config: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> EnvironmentCollectorInterface:
        custom_script = config.get("path", DEFAULT_CUSTOM_SCRIPT)
        if not isinstance(custom_script, str):
            message = f"Option `path` for environment collector `{cls.PLUGIN_NAME}` must be a string"
            raise TypeError(message)

        if not custom_script:
            message = f"Option `path` for environment collector `{cls.PLUGIN_NAME}` must not be empty if defined"
            raise ValueError(message)

        path = os.path.normpath(os.path.join(root, custom_script))
        if not os.path.isfile(path):
            message = f"Plugin script does not exist: {custom_script}"
            raise OSError(message)

        hook_class = load_plugin_from_script(
            path, custom_script, EnvironmentCollectorInterface, "environment_collector"
        )
        hook = hook_class(root, config, *args, **kwargs)

        # Always keep the name to avoid confusion
        hook.PLUGIN_NAME = cls.PLUGIN_NAME

        return hook



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import click

from hatch.config.constants import AppEnvVars

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Enter a shell within a project's environment")
@click.argument("env_name", required=False)
@click.option("--name")
@click.option("--path")
@click.pass_obj
def shell(app: Application, env_name: str | None, name: str, path: str):  # no cov
    """Enter a shell within a project's environment."""
    app.ensure_environment_plugin_dependencies()

    chosen_env = env_name or app.env
    if chosen_env == app.env_active:
        app.abort(f"Already in environment: {chosen_env}")

    for matrices in (app.project.config.matrices, app.project.config.internal_matrices):
        if chosen_env in matrices:
            app.display_error(f"Environment `{chosen_env}` defines a matrix, choose one of the following instead:\n")
            for generated_name in matrices[chosen_env]["envs"]:
                app.display_error(generated_name)

            app.abort()

    if not name:
        name = app.config.shell.name
    if not path:
        path = app.config.shell.path

    args = app.config.shell.args
    if not path:
        name, path = app.shell_data
        if not app.platform.windows:
            path, *args = app.platform.modules.shlex.split(path)

    with app.project.ensure_cwd():
        environment = app.project.get_environment(chosen_env)
        app.project.prepare_environment(environment, keep_env=bool(os.environ.get(AppEnvVars.KEEP_ENV)))

        first_run_indicator = app.cache_dir / "shell" / "first_run"
        if not first_run_indicator.is_file():
            app.display_waiting(
                "You are about to enter a new shell, exit as you usually would e.g. "
                "by typing `exit` or pressing `ctrl+d`..."
            )
            first_run_indicator.parent.ensure_dir_exists()
            first_run_indicator.touch()

        environment.enter_shell(name, path, args)



from hatch.template import File
from hatch.utils.fs import Path


class TestsPackageRoot(File):
    def __init__(
        self,
        template_config: dict,  # noqa: ARG002
        plugin_config: dict,  # noqa: ARG002
    ):
        super().__init__(Path("tests", "__init__.py"))



from __future__ import annotations


class EnvironmentCollectorInterface:
    """
    Example usage:

    ```python tab="plugin.py"
        from hatch.env.collectors.plugin.interface import EnvironmentCollectorInterface


        class SpecialEnvironmentCollector(EnvironmentCollectorInterface):
            PLUGIN_NAME = 'special'
            ...
    ```

    ```python tab="hooks.py"
        from hatchling.plugin import hookimpl

        from .plugin import SpecialEnvironmentCollector


        @hookimpl
        def hatch_register_environment_collector():
            return SpecialEnvironmentCollector
    ```
    """

    PLUGIN_NAME = ""
    """The name used for selection."""

    def __init__(self, root, config):
        self.__root = root
        self.__config = config

    @property
    def root(self):
        """
        The root of the project tree as a path-like object.
        """
        return self.__root

    @property
    def config(self) -> dict:
        """
        ```toml config-example
        [tool.hatch.env.collectors.<PLUGIN_NAME>]
        ```
        """
        return self.__config

    def get_initial_config(self) -> dict[str, dict]:  # noqa: PLR6301
        """
        Returns configuration for environments keyed by the environment or matrix name.
        """
        return {}

    def finalize_config(self, config: dict[str, dict]):
        """
        Finalizes configuration for environments keyed by the environment or matrix name. This will override
        any user-defined settings and any collectors that ran before this call.

        This is called before matrices are turned into concrete environments.
        """

    def finalize_environments(self, config: dict[str, dict]):
        """
        Finalizes configuration for environments keyed by the environment name. This will override
        any user-defined settings and any collectors that ran before this call.

        This is called after matrices are turned into concrete environments.
        """



class TemplateInterface:
    PLUGIN_NAME = ""
    PRIORITY = 100

    def __init__(self, plugin_config: dict, cache_dir, creation_time):
        self.plugin_config = plugin_config
        self.cache_dir = cache_dir
        self.creation_time = creation_time

    def initialize_config(self, config):
        """
        Allow modification of the configuration passed to every file for new projects
        before the list of files are determined.
        """

    def get_files(self, config):  # noqa: ARG002, PLR6301
        """Add to the list of files for new projects that are written to the file system."""
        return []

    def finalize_files(self, config, files):
        """Allow modification of files for new projects before they are written to the file system."""



import click


@click.command(short_help="Show information about the current environment")
@click.pass_obj
def status(app):
    """Show information about the current environment."""

    def display_pair(key, value, display_func=None, link=None):
        app.display_info("[", end="")
        app.display_success(key, end="")
        app.display_info("]", end="")
        app.display_info(" - ", end="")
        (display_func or app.display_info)(value, link=link)

    if app.project.root is None:
        if app.project.chosen_name is None:
            display_pair("Project", "<no project detected>", app.display_warning)
        else:
            display_pair("Project", f"{app.project.chosen_name} (not a project)", app.display_warning)
    elif app.project.chosen_name is None:
        display_pair("Project", f"{app.project.root.name} (current directory)")
    else:
        display_pair("Project", app.project.chosen_name)

    display_pair("Location", str(app.project.location))
    display_pair("Config", str(app.config_file.path), link=f"file:///{app.config_file.path}")



from hatch.index.errors import ArtifactMetadataError

MULTIPLE_USE_METADATA_FIELDS = {
    "classifier",
    "dynamic",
    "license_file",
    "obsoletes_dist",
    "platform",
    "project_url",
    "provides_dist",
    "provides_extra",
    "requires_dist",
    "requires_external",
    "supported_platform",
}
RENAMED_METADATA_FIELDS = {"classifier": "classifiers", "project_url": "project_urls"}


def get_wheel_form_data(artifact):
    import zipfile

    from packaging.tags import parse_tag

    with zipfile.ZipFile(str(artifact), "r") as zip_archive:
        for path in zip_archive.namelist():
            root = path.split("/", 1)[0]
            if root.endswith(".dist-info"):
                dist_info_dir = root
                break
        else:  # no cov
            message = f"Could not find the `.dist-info` directory in wheel: {artifact}"
            raise ArtifactMetadataError(message)

        try:
            with zip_archive.open(f"{dist_info_dir}/METADATA") as zip_file:
                metadata_file_contents = zip_file.read().decode("utf-8")
        except KeyError:  # no cov
            message = f"Could not find a `METADATA` file in the `{dist_info_dir}` directory"
            raise ArtifactMetadataError(message) from None
        else:
            data = parse_headers(metadata_file_contents)

    data["filetype"] = "bdist_wheel"

    # Examples:
    # cryptography-3.4.7-pp37-pypy37_pp73-manylinux2014_x86_64.whl -> pp37
    # hatchling-1rc1-py2.py3-none-any.whl -> py2.py3
    tag_component = "-".join(artifact.stem.split("-")[-3:])
    data["pyversion"] = ".".join(sorted({tag.interpreter for tag in parse_tag(tag_component)}))

    return data


def get_sdist_form_data(artifact):
    import tarfile

    with tarfile.open(str(artifact), "r:gz") as tar_archive:
        pkg_info_dir_parts = []
        for tar_info in tar_archive:
            if tar_info.isfile():
                pkg_info_dir_parts.append(tar_info.name.split("/", 1)[0])
                break
        else:  # no cov
            message = f"Could not find any files in sdist: {artifact}"
            raise ArtifactMetadataError(message)

        pkg_info_dir_parts.append("PKG-INFO")
        pkg_info_path = "/".join(pkg_info_dir_parts)
        try:
            with tar_archive.extractfile(pkg_info_path) as tar_file:
                metadata_file_contents = tar_file.read().decode("utf-8")
        except KeyError:  # no cov
            message = f"Could not find file: {pkg_info_path}"
            raise ArtifactMetadataError(message) from None
        else:
            data = parse_headers(metadata_file_contents)

    data["filetype"] = "sdist"
    data["pyversion"] = "source"

    return data


def parse_headers(metadata_file_contents):
    import email

    message = email.message_from_string(metadata_file_contents)

    headers = {"description": message.get_payload()}

    for header, value in message.items():
        normalized_header = header.lower().replace("-", "_")
        header_name = RENAMED_METADATA_FIELDS.get(normalized_header, normalized_header)

        if normalized_header in MULTIPLE_USE_METADATA_FIELDS:
            if header_name in headers:
                headers[header_name].append(value)
            else:
                headers[header_name] = [value]
        else:
            headers[header_name] = value

    return headers



from __future__ import annotations

import re
from copy import deepcopy
from functools import cached_property
from itertools import product
from os import environ
from typing import TYPE_CHECKING, Any

from hatch.env.utils import ensure_valid_environment
from hatch.project.constants import DEFAULT_BUILD_DIRECTORY, BuildEnvVars
from hatch.project.env import apply_overrides
from hatch.project.utils import format_script_commands, parse_script_command

if TYPE_CHECKING:
    from hatch.dep.core import Dependency


class ProjectConfig:
    def __init__(self, root, config, plugin_manager=None):
        self.root = root
        self.config = config
        self.plugin_manager = plugin_manager

        self._matrices = None
        self._env = None
        self._env_requires_complex = None
        self._env_requires = None
        self._env_collectors = None
        self._envs = None
        self._internal_envs = None
        self._internal_matrices = None
        self._matrix_variables = None
        self._publish = None
        self._scripts = None
        self._cached_env_overrides = {}

    @cached_property
    def build(self):
        config = self.config.get("build", {})
        if not isinstance(config, dict):
            message = "Field `tool.hatch.build` must be a table"
            raise TypeError(message)

        return BuildConfig(config)

    @cached_property
    def lock_envs(self) -> bool:
        lock_envs = self.config.get("lock-envs", False)
        if not isinstance(lock_envs, bool):
            message = "Field `tool.hatch.lock-envs` must be a boolean"
            raise TypeError(message)

        return lock_envs

    @cached_property
    def locker(self) -> str | None:
        locker = self.config.get("locker")
        if locker is None:
            return None
        if not isinstance(locker, str):
            message = "Field `tool.hatch.locker` must be a string"
            raise TypeError(message)

        return locker

    @property
    def env(self):
        if self._env is None:
            config = self.config.get("env", {})
            if not isinstance(config, dict):
                message = "Field `tool.hatch.env` must be a table"
                raise TypeError(message)

            self._env = config

        return self._env

    @property
    def env_requires_complex(self) -> list[Dependency]:
        if self._env_requires_complex is None:
            from hatch.dep.core import Dependency, InvalidDependencyError

            requires = self.env.get("requires", [])
            if not isinstance(requires, list):
                message = "Field `tool.hatch.env.requires` must be an array"
                raise TypeError(message)

            requires_complex = []

            for i, entry in enumerate(requires, 1):
                if not isinstance(entry, str):
                    message = f"Requirement #{i} in `tool.hatch.env.requires` must be a string"
                    raise TypeError(message)

                try:
                    requires_complex.append(Dependency(entry))
                except InvalidDependencyError as e:
                    message = f"Requirement #{i} in `tool.hatch.env.requires` is invalid: {e}"
                    raise ValueError(message) from None

            self._env_requires_complex = requires_complex

        return self._env_requires_complex

    @property
    def env_requires(self):
        if self._env_requires is None:
            self._env_requires = [str(r) for r in self.env_requires_complex]

        return self._env_requires

    @property
    def env_collectors(self):
        if self._env_collectors is None:
            collectors = self.env.get("collectors", {})
            if not isinstance(collectors, dict):
                message = "Field `tool.hatch.env.collectors` must be a table"
                raise TypeError(message)

            final_config = {"default": {}}
            for collector, config in collectors.items():
                if not isinstance(config, dict):
                    message = f"Field `tool.hatch.env.collectors.{collector}` must be a table"
                    raise TypeError(message)

                final_config[collector] = config

            self._env_collectors = final_config

        return self._env_collectors

    @property
    def matrices(self):
        if self._matrices is None:
            _ = self.envs

        return self._matrices

    @property
    def matrix_variables(self):
        if self._matrix_variables is None:
            _ = self.envs

        return self._matrix_variables

    @property
    def internal_envs(self):
        if self._internal_envs is None:
            _ = self.envs

        return self._internal_envs

    @property
    def internal_matrices(self):
        if self._internal_matrices is None:
            _ = self.envs

        return self._internal_matrices

    @property
    def envs(self):
        from hatch.env.internal import get_internal_env_config
        from hatch.utils.platform import get_platform_name

        if self._envs is None:
            env_config = self.config.get("envs", {})
            if not isinstance(env_config, dict):
                message = "Field `tool.hatch.envs` must be a table"
                raise TypeError(message)

            config = {}
            environment_collectors = []

            for collector, collector_config in self.env_collectors.items():
                collector_class = self.plugin_manager.environment_collector.get(collector)
                if collector_class is None:
                    message = f"Unknown environment collector: {collector}"
                    raise ValueError(message)

                environment_collector = collector_class(self.root, collector_config)
                environment_collectors.append(environment_collector)

                for env_name, data in environment_collector.get_initial_config().items():
                    config.setdefault(env_name, data)

            for env_name, data in env_config.items():
                if not isinstance(data, dict):
                    message = f"Field `tool.hatch.envs.{env_name}` must be a table"
                    raise TypeError(message)

                config.setdefault(env_name, {}).update(data)

            for environment_collector in environment_collectors:
                environment_collector.finalize_config(config)

            # Prevent plugins from removing the default environment
            ensure_valid_environment(config.setdefault("default", {}))

            seen = set()
            active = []
            for env_name, data in config.items():
                _populate_default_env_values(env_name, data, config, seen, active)

            current_platform = get_platform_name()
            all_matrices = {}
            generated_envs = {}
            final_config = {}
            cached_overrides = {}
            for env_name, raw_initial_config in config.items():
                current_cached_overrides = cached_overrides[env_name] = {
                    "platform": [],
                    "env": [],
                    "matrix": [],
                    "name": [],
                }

                # Only shallow copying is necessary since we just want to modify keys
                initial_config = raw_initial_config.copy()

                matrix_name_format = initial_config.pop("matrix-name-format", "{value}")
                if not isinstance(matrix_name_format, str):
                    message = f"Field `tool.hatch.envs.{env_name}.matrix-name-format` must be a string"
                    raise TypeError(message)

                if "{value}" not in matrix_name_format:
                    message = (
                        f"Field `tool.hatch.envs.{env_name}.matrix-name-format` must "
                        f"contain at least the `{{value}}` placeholder"
                    )
                    raise ValueError(message)

                overrides = initial_config.pop("overrides", {})
                if not isinstance(overrides, dict):
                    message = f"Field `tool.hatch.envs.{env_name}.overrides` must be a table"
                    raise TypeError(message)

                # Apply any configuration based on the current platform
                platform_overrides = overrides.get("platform", {})
                if not isinstance(platform_overrides, dict):
                    message = f"Field `tool.hatch.envs.{env_name}.overrides.platform` must be a table"
                    raise TypeError(message)

                for platform, options in platform_overrides.items():
                    if not isinstance(options, dict):
                        message = f"Field `tool.hatch.envs.{env_name}.overrides.platform.{platform}` must be a table"
                        raise TypeError(message)

                    if platform != current_platform:
                        continue

                    apply_overrides(env_name, "platform", platform, current_platform, options, initial_config)
                    current_cached_overrides["platform"].append((platform, current_platform, options))

                # Apply any configuration based on environment variables
                env_var_overrides = overrides.get("env", {})
                if not isinstance(env_var_overrides, dict):
                    message = f"Field `tool.hatch.envs.{env_name}.overrides.env` must be a table"
                    raise TypeError(message)

                for env_var, options in env_var_overrides.items():
                    if not isinstance(options, dict):
                        message = f"Field `tool.hatch.envs.{env_name}.overrides.env.{env_var}` must be a table"
                        raise TypeError(message)

                    if env_var not in environ:
                        continue

                    apply_overrides(env_name, "env", env_var, environ[env_var], options, initial_config)
                    current_cached_overrides["env"].append((env_var, environ[env_var], options))

                if "matrix" not in initial_config:
                    final_config[env_name] = initial_config
                    continue

                matrices = initial_config.pop("matrix")
                if not isinstance(matrices, list):
                    message = f"Field `tool.hatch.envs.{env_name}.matrix` must be an array"
                    raise TypeError(message)

                matrix_overrides = overrides.get("matrix", {})
                if not isinstance(matrix_overrides, dict):
                    message = f"Field `tool.hatch.envs.{env_name}.overrides.matrix` must be a table"
                    raise TypeError(message)

                name_overrides = overrides.get("name", {})
                if not isinstance(name_overrides, dict):
                    message = f"Field `tool.hatch.envs.{env_name}.overrides.name` must be a table"
                    raise TypeError(message)

                matrix_data = all_matrices[env_name] = {"config": deepcopy(initial_config)}
                all_envs = matrix_data["envs"] = {}
                for i, raw_matrix in enumerate(matrices, 1):
                    matrix = raw_matrix
                    if not isinstance(matrix, dict):
                        message = f"Entry #{i} in field `tool.hatch.envs.{env_name}.matrix` must be a table"
                        raise TypeError(message)

                    if not matrix:
                        message = f"Matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` cannot be empty"
                        raise ValueError(message)

                    for j, (variable, values) in enumerate(matrix.items(), 1):
                        if not variable:
                            message = (
                                f"Variable #{j} in matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` "
                                f"cannot be an empty string"
                            )
                            raise ValueError(message)

                        if not isinstance(values, list):
                            message = (
                                f"Variable `{variable}` in matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` "
                                f"must be an array"
                            )
                            raise TypeError(message)

                        if not values:
                            message = (
                                f"Variable `{variable}` in matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` "
                                f"cannot be empty"
                            )
                            raise ValueError(message)

                        existing_values = set()
                        for k, value in enumerate(values, 1):
                            if not isinstance(value, str):
                                message = (
                                    f"Value #{k} of variable `{variable}` in matrix #{i} in field "
                                    f"`tool.hatch.envs.{env_name}.matrix` must be a string"
                                )
                                raise TypeError(message)

                            if not value:
                                message = (
                                    f"Value #{k} of variable `{variable}` in matrix #{i} in field "
                                    f"`tool.hatch.envs.{env_name}.matrix` cannot be an empty string"
                                )
                                raise ValueError(message)

                            if value in existing_values:
                                message = (
                                    f"Value #{k} of variable `{variable}` in matrix #{i} in field "
                                    f"`tool.hatch.envs.{env_name}.matrix` is a duplicate"
                                )
                                raise ValueError(message)

                            existing_values.add(value)

                    variables = {}

                    # Ensure that any Python variable comes first
                    python_selected = False
                    for variable in ("py", "python"):
                        if variable in matrix:
                            if python_selected:
                                message = (
                                    f"Matrix #{i} in field `tool.hatch.envs.{env_name}.matrix` "
                                    f"cannot contain both `py` and `python` variables"
                                )
                                raise ValueError(message)
                            python_selected = True

                            # Only shallow copying is necessary since we just want to remove a key
                            matrix = matrix.copy()
                            variables[variable] = matrix.pop(variable)

                    variables.update(matrix)

                    for result in product(*variables.values()):
                        # Make a value mapping for easy referencing
                        variable_values = dict(zip(variables, result, strict=False))

                        # Create the environment's initial configuration
                        new_config = deepcopy(initial_config)

                        cached_matrix_overrides = []

                        # Apply any configuration based on matrix variables
                        for variable, options in matrix_overrides.items():
                            if not isinstance(options, dict):
                                message = (
                                    f"Field `tool.hatch.envs.{env_name}.overrides.matrix.{variable}` must be a table"
                                )
                                raise TypeError(message)

                            if variable not in variables:
                                continue

                            apply_overrides(
                                env_name, "matrix", variable, variable_values[variable], options, new_config
                            )
                            cached_matrix_overrides.append((variable, variable_values[variable], options))

                        # Construct the environment name
                        final_matrix_name_format = new_config.pop("matrix-name-format", matrix_name_format)
                        env_name_parts = []
                        for j, (variable, value) in enumerate(variable_values.items()):
                            if j == 0 and python_selected:
                                new_config["python"] = value
                                env_name_parts.append(value if value.startswith("py") else f"py{value}")
                            else:
                                env_name_parts.append(final_matrix_name_format.format(variable=variable, value=value))

                        new_env_name = "-".join(env_name_parts)

                        cached_name_overrides = []

                        # Apply any configuration based on the final name, minus the prefix for non-default environments
                        for pattern, options in name_overrides.items():
                            if not isinstance(options, dict):
                                message = f"Field `tool.hatch.envs.{env_name}.overrides.name.{pattern}` must be a table"
                                raise TypeError(message)

                            if not re.search(pattern, new_env_name):
                                continue

                            apply_overrides(env_name, "name", pattern, new_env_name, options, new_config)
                            cached_name_overrides.append((pattern, new_env_name, options))

                        if env_name != "default":
                            new_env_name = f"{env_name}.{new_env_name}"

                        # Save the generated environment
                        final_config[new_env_name] = new_config
                        cached_overrides[new_env_name] = {
                            "platform": current_cached_overrides["platform"],
                            "env": current_cached_overrides["env"],
                            "matrix": cached_matrix_overrides,
                            "name": cached_name_overrides,
                        }
                        all_envs[new_env_name] = variable_values
                        if "py" in variable_values:
                            all_envs[new_env_name] = {"python": variable_values.pop("py"), **variable_values}

                # Remove the root matrix generator
                del cached_overrides[env_name]

                # Save the variables used to generate the environments
                generated_envs.update(all_envs)

            for environment_collector in environment_collectors:
                environment_collector.finalize_environments(final_config)

            self._matrices = all_matrices
            self._internal_matrices = {}
            self._envs = final_config
            self._matrix_variables = generated_envs
            self._cached_env_overrides.update(cached_overrides)

            # Extract the internal environments
            self._internal_envs = {}
            for internal_name in get_internal_env_config():
                try:
                    self._internal_envs[internal_name] = self._envs.pop(internal_name)
                # Matrix
                except KeyError:
                    self._internal_matrices[internal_name] = self._matrices.pop(internal_name)
                    for env_name in [env_name for env_name in self._envs if env_name.startswith(f"{internal_name}.")]:
                        self._internal_envs[env_name] = self._envs.pop(env_name)

        return self._envs

    @property
    def publish(self):
        if self._publish is None:
            config = self.config.get("publish", {})
            if not isinstance(config, dict):
                message = "Field `tool.hatch.publish` must be a table"
                raise TypeError(message)

            for publisher, data in config.items():
                if not isinstance(data, dict):
                    message = f"Field `tool.hatch.publish.{publisher}` must be a table"
                    raise TypeError(message)

            self._publish = config

        return self._publish

    @property
    def scripts(self):
        if self._scripts is None:
            script_config = self.config.get("scripts", {})
            if not isinstance(script_config, dict):
                message = "Field `tool.hatch.scripts` must be a table"
                raise TypeError(message)

            config = {}

            for name, data in script_config.items():
                if " " in name:
                    message = f"Script name `{name}` in field `tool.hatch.scripts` must not contain spaces"
                    raise ValueError(message)

                commands = []

                if isinstance(data, str):
                    commands.append(data)
                elif isinstance(data, list):
                    for i, command in enumerate(data, 1):
                        if not isinstance(command, str):
                            message = f"Command #{i} in field `tool.hatch.scripts.{name}` must be a string"
                            raise TypeError(message)

                        commands.append(command)
                else:
                    message = f"Field `tool.hatch.scripts.{name}` must be a string or an array of strings"
                    raise TypeError(message)

                config[name] = commands

            seen = {}
            active = []
            for script_name, commands in config.items():
                commands[:] = expand_script_commands(script_name, commands, config, seen, active)

            self._scripts = config

        return self._scripts

    def finalize_env_overrides(self, option_types):
        # We lazily apply overrides because we need type information potentially defined by
        # environment plugins for their options
        if not self._cached_env_overrides:
            return

        for environments in (self.envs, self.internal_envs):
            for env_name, config in environments.items():
                for override_name, data in self._cached_env_overrides.get(env_name, {}).items():
                    for condition, condition_value, options in data:
                        apply_overrides(
                            env_name, override_name, condition, condition_value, options, config, option_types
                        )

        self._cached_env_overrides.clear()


class BuildConfig:
    def __init__(self, config: dict[str, Any]) -> None:
        self.__config = config
        self.__targets: dict[str, BuildTargetConfig] = config

    @cached_property
    def directory(self) -> str:
        directory = self.__config.get("directory", DEFAULT_BUILD_DIRECTORY)
        if not isinstance(directory, str):
            message = "Field `tool.hatch.build.directory` must be a string"
            raise TypeError(message)

        return directory

    @cached_property
    def dependencies(self) -> list[str]:
        dependencies: list[str] = self.__config.get("dependencies", [])
        if not isinstance(dependencies, list):
            message = "Field `tool.hatch.build.dependencies` must be an array"
            raise TypeError(message)

        for i, dependency in enumerate(dependencies, 1):
            if not isinstance(dependency, str):
                message = f"Dependency #{i} in field `tool.hatch.build.dependencies` must be a string"
                raise TypeError(message)

        return list(dependencies)

    @cached_property
    def hook_config(self) -> dict[str, dict[str, Any]]:
        hook_config: dict[str, dict[str, Any]] = self.__config.get("hooks", {})
        if not isinstance(hook_config, dict):
            message = "Field `tool.hatch.build.hooks` must be a table"
            raise TypeError(message)

        for hook_name, config in hook_config.items():
            if not isinstance(config, dict):
                message = f"Field `tool.hatch.build.hooks.{hook_name}` must be a table"
                raise TypeError(message)

        return finalize_hook_config(hook_config)

    @cached_property
    def __target_config(self) -> dict[str, Any]:
        config = self.__config.get("targets", {})
        if not isinstance(config, dict):
            message = "Field `tool.hatch.build.targets` must be a table"
            raise TypeError(message)

        return config

    def target(self, target: str) -> BuildTargetConfig:
        if target in self.__targets:
            return self.__targets[target]

        config = self.__target_config.get(target, {})
        if not isinstance(config, dict):
            message = f"Field `tool.hatch.build.targets.{target}` must be a table"
            raise TypeError(message)

        target_config = BuildTargetConfig(target, config, self)
        self.__targets[target] = target_config
        return target_config


class BuildTargetConfig:
    def __init__(self, name: str, config: dict[str, Any], global_config: BuildConfig) -> None:
        self.__name = name
        self.__config = config
        self.__global_config = global_config

    @cached_property
    def directory(self) -> str:
        directory = self.__config.get("directory", self.__global_config.directory)
        if not isinstance(directory, str):
            message = f"Field `tool.hatch.build.targets.{self.__name}.directory` must be a string"
            raise TypeError(message)

        return directory

    @cached_property
    def dependencies(self) -> list[str]:
        dependencies: list[str] = self.__config.get("dependencies", [])
        if not isinstance(dependencies, list):
            message = f"Field `tool.hatch.build.targets.{self.__name}.dependencies` must be an array"
            raise TypeError(message)

        for i, dependency in enumerate(dependencies, 1):
            if not isinstance(dependency, str):
                message = (
                    f"Dependency #{i} in field `tool.hatch.build.targets.{self.__name}.dependencies` must be a string"
                )
                raise TypeError(message)

        all_dependencies = list(self.__global_config.dependencies)
        all_dependencies.extend(dependencies)
        return all_dependencies

    @cached_property
    def hook_config(self) -> dict[str, dict[str, Any]]:
        hook_config: dict[str, dict[str, Any]] = self.__config.get("hooks", {})
        if not isinstance(hook_config, dict):
            message = f"Field `tool.hatch.build.targets.{self.__name}.hooks` must be a table"
            raise TypeError(message)

        for hook_name, config in hook_config.items():
            if not isinstance(config, dict):
                message = f"Field `tool.hatch.build.targets.{self.__name}.hooks.{hook_name}` must be a table"
                raise TypeError(message)

        config = self.__global_config.hook_config.copy()
        config.update(hook_config)
        return finalize_hook_config(config)


def expand_script_commands(script_name, commands, config, seen, active):
    if script_name in seen:
        return seen[script_name]

    if script_name in active:
        active.append(script_name)

        message = f"Circular expansion detected for field `tool.hatch.scripts`: {' -> '.join(active)}"
        raise ValueError(message)

    active.append(script_name)

    expanded_commands = []

    for command in commands:
        possible_script, args, ignore_exit_code = parse_script_command(command)

        if possible_script in config:
            expanded_commands.extend(
                format_script_commands(
                    commands=expand_script_commands(possible_script, config[possible_script], config, seen, active),
                    args=args,
                    ignore_exit_code=ignore_exit_code,
                )
            )
        else:
            expanded_commands.append(command)

    seen[script_name] = expanded_commands
    active.pop()

    return expanded_commands


def _populate_default_env_values(env_name, data, config, seen, active):
    if env_name in seen:
        return

    if data.pop("detached", False):
        data["template"] = env_name
        data["skip-install"] = True

    template_name = data.pop("template", "default")
    if template_name not in config:
        message = f"Field `tool.hatch.envs.{env_name}.template` refers to an unknown environment `{template_name}`"
        raise ValueError(message)

    if env_name in active:
        active.append(env_name)

        message = f"Circular inheritance detected for field `tool.hatch.envs.*.template`: {' -> '.join(active)}"
        raise ValueError(message)

    if template_name == env_name:
        ensure_valid_environment(data)
        seen.add(env_name)
        return

    active.append(env_name)

    template_config = config[template_name]
    _populate_default_env_values(template_name, template_config, config, seen, active)

    for key, value in template_config.items():
        if key == "matrix":
            continue

        if key == "scripts":
            scripts = data["scripts"] if "scripts" in data else data.setdefault("scripts", {})
            for script, commands in value.items():
                scripts.setdefault(script, commands)
        else:
            data.setdefault(key, value)

    seen.add(env_name)
    active.pop()


def finalize_hook_config(hook_config: dict[str, dict[str, Any]]) -> dict[str, dict[str, Any]]:
    if env_var_enabled(BuildEnvVars.NO_HOOKS):
        return {}

    all_hooks_enabled = env_var_enabled(BuildEnvVars.HOOKS_ENABLE)
    final_hook_config: dict[str, dict[str, Any]] = {
        hook_name: config
        for hook_name, config in hook_config.items()
        if (
            all_hooks_enabled
            or config.get("enable-by-default", True)
            or env_var_enabled(f"{BuildEnvVars.HOOK_ENABLE_PREFIX}{hook_name.upper()}")
        )
    }
    return final_hook_config


def env_var_enabled(env_var: str, *, default: bool = False) -> bool:
    if env_var in environ:
        return environ[env_var] in {"1", "true"}

    return default



class AppEnvVars:
    ENV = "HATCH_ENV"
    ENV_ACTIVE = "HATCH_ENV_ACTIVE"
    ENV_OPTION_PREFIX = "HATCH_ENV_TYPE_"
    QUIET = "HATCH_QUIET"
    VERBOSE = "HATCH_VERBOSE"
    INTERACTIVE = "HATCH_INTERACTIVE"
    PYTHON = "HATCH_PYTHON"
    # https://no-color.org
    NO_COLOR = "NO_COLOR"
    FORCE_COLOR = "FORCE_COLOR"
    KEEP_ENV = "HATCH_KEEP_ENV"


class ConfigEnvVars:
    PROJECT = "HATCH_PROJECT"
    DATA = "HATCH_DATA_DIR"
    CACHE = "HATCH_CACHE_DIR"
    CONFIG = "HATCH_CONFIG"


class PublishEnvVars:
    USER = "HATCH_INDEX_USER"
    AUTH = "HATCH_INDEX_AUTH"
    REPO = "HATCH_INDEX_REPO"
    CA_CERT = "HATCH_INDEX_CA_CERT"
    CLIENT_CERT = "HATCH_INDEX_CLIENT_CERT"
    CLIENT_KEY = "HATCH_INDEX_CLIENT_KEY"
    PUBLISHER = "HATCH_PUBLISHER"
    OPTIONS = "HATCH_PUBLISHER_OPTIONS"


class PythonEnvVars:
    CUSTOM_SOURCE_PREFIX = "HATCH_PYTHON_CUSTOM_SOURCE_"
    CUSTOM_PATH_PREFIX = "HATCH_PYTHON_CUSTOM_PATH_"
    CUSTOM_VERSION_PREFIX = "HATCH_PYTHON_CUSTOM_VERSION_"


class VersionEnvVars:
    VALIDATE_BUMP = "HATCH_VERSION_VALIDATE_BUMP"



from __future__ import annotations

from abc import ABC, abstractmethod


class PublisherInterface(ABC):
    """
    Example usage:

    ```python tab="plugin.py"
        from hatch.publish.plugin.interface import PublisherInterface


        class SpecialPublisher(PublisherInterface):
            PLUGIN_NAME = 'special'
            ...
    ```

    ```python tab="hooks.py"
        from hatchling.plugin import hookimpl

        from .plugin import SpecialPublisher


        @hookimpl
        def hatch_register_publisher():
            return SpecialPublisher
    ```
    """

    PLUGIN_NAME = ""
    """The name used for selection."""

    def __init__(self, app, root, cache_dir, project_config, plugin_config):
        self.__app = app
        self.__root = root
        self.__cache_dir = cache_dir
        self.__project_config = project_config
        self.__plugin_config = plugin_config

        self.__disable = None

    @property
    def app(self):
        """
        An instance of [Application](../utilities.md#hatchling.bridge.app.Application).
        """
        return self.__app

    @property
    def root(self):
        """
        The root of the project tree as a path-like object.
        """
        return self.__root

    @property
    def cache_dir(self):
        """
        The directory reserved exclusively for this plugin as a path-like object.
        """
        return self.__cache_dir

    @property
    def project_config(self) -> dict:
        """
        ```toml config-example
        [tool.hatch.publish.<PLUGIN_NAME>]
        ```
        """
        return self.__project_config

    @property
    def plugin_config(self) -> dict:
        """
        This is defined in Hatch's [config file](../../config/hatch.md).

        ```toml tab="config.toml"
        [publish.<PLUGIN_NAME>]
        ```
        """
        return self.__plugin_config

    @property
    def disable(self):
        """
        Whether this plugin is disabled, thus requiring confirmation when publishing. Local
        [project configuration](reference.md#hatch.publish.plugin.interface.PublisherInterface.project_config)
        takes precedence over global
        [plugin configuration](reference.md#hatch.publish.plugin.interface.PublisherInterface.plugin_config).
        """
        if self.__disable is None:
            if "disable" in self.project_config:
                disable = self.project_config["disable"]
                if not isinstance(disable, bool):
                    message = f"Field `tool.hatch.publish.{self.PLUGIN_NAME}.disable` must be a boolean"
                    raise TypeError(message)
            else:
                disable = self.plugin_config.get("disable", False)
                if not isinstance(disable, bool):
                    message = f"Global plugin configuration `publish.{self.PLUGIN_NAME}.disable` must be a boolean"
                    raise TypeError(message)

            self.__disable = disable

        return self.__disable

    @abstractmethod
    def publish(self, artifacts: list[str], options: dict):
        """
        :material-align-horizontal-left: **REQUIRED** :material-align-horizontal-right:

        This is called directly by the [`publish`](../../cli/reference.md#hatch-publish) command
        with the arguments and options it receives.
        """
